local B, L, T = unpack(select(2, ...))

T.LABEL = {
    ['bank'] = L[10001],
    ['reagent'] = L[10002],
    ['void'] = L[10003],
    ['guild'] = L[10004],
    ['player'] = L[10010],
    ['equip'] = L[10018]
}

T.TOOLTIPLABEL = {
    ['bank'] = L[10013],
    ['reagent'] = L[10014],
    ['void'] = L[10016],
    ['guild'] = L[10017],
    ['player'] = L[10015],
    ['equipped'] = L[10012]
}

--[[//////////////////////////////////////////////////

    FUNCTIONS TRIGGERED VIA XML bankswatch.xml

----\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\]]

    function BW_OnLoad(self)

        ----- REGISTER FOR EVENTS
            self:RegisterEvent('BANKFRAME_OPENED')
            self:RegisterEvent('BANKFRAME_CLOSED')
            self:RegisterEvent('PLAYERBANKSLOTS_CHANGED')
            self:RegisterEvent('PLAYERREAGENTBANKSLOTS_CHANGED')
            self:RegisterEvent('VOID_STORAGE_OPEN')
            self:RegisterEvent('VOID_STORAGE_CLOSE')
            self:RegisterEvent('GUILDBANKFRAME_OPENED')
            self:RegisterEvent('GUILDBANK_UPDATE_TABS')
            self:RegisterEvent('PLAYER_ENTERING_WORLD')
            self:RegisterEvent('VARIABLES_LOADED')
            self:RegisterEvent('PLAYER_REGEN_DISABLED')
            self:RegisterEvent('GET_ITEM_INFO_RECEIVED')
            self:RegisterEvent('PLAYER_EQUIPMENT_CHANGED')
            self:RegisterEvent('PLAYERBANKBAGSLOTS_CHANGED')

            StackSplitFrame:SetFrameStrata("DIALOG") -- For some reason, this frame is not set as a dialog

        DEFAULT_CHAT_FRAME:AddMessage(T.MSGC.DEFAULT..L[10000])

        ------ SLASH COMMANDS
            SLASH_BW1 = "/BW"
            SLASH_BW2 = "/bw"
            SLASH_BW3 = "/Bw"
            SLASH_BW4 = "/bW"
            SLASH_BW5 = "/Bankswatch"
            SLASH_BW6 = "/BanksWatch"
            SLASH_BW7 = "/BANKSWATCH"
            SLASH_BW8 = "/bankswatch"

            SlashCmdList['BW'] = B.Main

            B.CreateFrames()
            table.insert(UISpecialFrames, T.F.mainframe.bank:GetName())
            table.insert(UISpecialFrames, T.F.mainframe.reagent:GetName())
            table.insert(UISpecialFrames, T.F.mainframe.void:GetName())
            table.insert(UISpecialFrames, T.F.mainframe.guild:GetName())
            table.insert(UISpecialFrames, T.F.mainframe.search:GetName())
            table.insert(UISpecialFrames, T.F.mainframe.character:GetName())

    end

    local bag_to_update = nil
    function BW_OnEvent(self, event, ...)

        if event == "VARIABLES_LOADED" then
            B.print(T.MSGC.EVENT.."Variables loaded. Initializing.", 2)
            B.print(T.MSGC.INFO.."Running version: "..GetAddOnMetadata("BanksWatch", "Version"), 1)

            if BWsavedtable then
                T.BUTTON_HEIGHT = BWsavedtable['iconsize'] or 32
                T.ROW = BWsavedtable['iconrow'] or 12

                if BWsavedtable['minimize'] then T.F.mainframe.togglebutton:Show() end
                if not BWsavedtable['bankbagstoggle'] then T.F.mainframe.bank.bags:Hide() end
                if not BWsavedtable['playerbagstoggle'] then T.F.mainframe.character.bags:Hide() end
            end

            B.I.Main.Bank.Buttons()
            B.I.Main.Reagent.Buttons()
            B.I.Main.Character.Buttons()
            B.I.Main.Void.Buttons()
            B.I.Main.Guild.Buttons()

        elseif event == "GET_ITEM_INFO_RECEIVED" then
            local itemid = ...
            if T.BW.lookup[itemid] then T.BW.lookup[itemid] = nil end

            if T.BW.lookuprequest and not next(T.BW.lookup) then -- Try to minimize the amount of updates
                T.BW.lookuprequest = false
                B.BankUpdate()
                B.ReagentUpdate()
                B.VoidUpdate()
                B.GuildUpdate()
                B.PlayerUpdate()
                B.PlayerEquipmentUpdate()
            end

        elseif event == "PLAYER_ENTERING_WORLD" then
            -- When loading, BAG_UPDATE triggers, but I only want when player modifies their bank bags
            if not T.BW.loaded then -- First time loading character
                T.BW.loaded = true
                self:RegisterEvent('PLAYER_REGEN_DISABLED')
                self:RegisterEvent('BAG_UPDATE')

                B.Initialize() -- Load or create character
                B.C.Create() -- Config
                B.SetFrameLocation()

                B.PlayerBagsScan()
                B.PlayerEquipmentScan()
                B.PlayerEquipmentUpdate()

                -- Cache items here
                B.CacheMainItems()
                B.CacheAltItems() -- Runs only once
            end

        elseif event == "PLAYERBANKSLOTS_CHANGED" or event == "PLAYERBANKBAGSLOTS_CHANGED" then
            -- Only for generic bank bag
            local arg = ...

            if arg then -- Sometimes, arg can be nil when getting in combat and set is switched
                if arg <= NUM_BANKGENERIC_SLOTS then -- Generic bank modified
                    B.GenericBankSlotScan(arg)
                    B.CacheMainItems('bank')
                elseif arg > NUM_BANKGENERIC_SLOTS then -- Bank bag modified, will trigger BAG_UPDATE
                    bag_to_update = arg - NUM_BANKGENERIC_SLOTS + 4
                end
            end

        elseif event == "PLAYER_EQUIPMENT_CHANGED" then
            B.PlayerEquipmentSlotScan(...)
            B.CacheMainItems('equipped')

        elseif event == "BAG_UPDATE" then
            local bag = ...
            --B.print("Bag to update: "..bag)

            if bag == 0 or bag <= NUM_BAG_SLOTS then
                -- Only run when bags are not open since magic doesn't happen unless bags are open.
                if not IsBagOpen(bag) then B.PlayerBagScan(bag); B.CacheMainItems('player') end

            elseif bag_to_update == bag then
                B.RefreshBankBags()
                B.BankBagScan(bag)
            end

        elseif event == "VOID_STORAGE_OPEN" then
            T.ISVOIDOPEN = true

        elseif event == "VOID_STORAGE_CLOSE" then
            T.ISVOIDOPEN = false
            B.CacheMainItems('void')

        elseif event == "BANKFRAME_OPENED" then
            T.ISBANKOPEN = true
            B.RefreshBankBags()

        elseif event == "BANKFRAME_CLOSED" then
            T.ISBANKOPEN = false
            B.CacheMainItems('bank')
            B.CacheMainItems('reagent')

        elseif event == "GUILDBANKFRAME_OPENED" or event == "GUILDBANK_UPDATE_TABS" then
            if event == "GUILDBANKFRAME_OPENED" then self:RegisterEvent("GUILDBANKFRAME_CLOSED") end
            T.ISGUILDOPEN = true
            B.RefreshGuildTab()

        elseif event == "GUILDBANKFRAME_CLOSED" then
            self:UnregisterEvent(event)
            T.ISGUILDOPEN = false
            B.CacheMainItems('guild')

        elseif event == "PLAYER_REGEN_DISABLED" then
            -- Entering combat, hide any open bags
            B.HideAll()

        elseif event == "PLAYERREAGENTBANKSLOTS_CHANGED" then
            B.ReagentSlotScan(...)
            B.CacheMainItems('reagent')
        end
    end

    function BW_ToggleFrame() -- Bindings.xml
        if T.F.toggleframe:IsVisible() then
            T.F.toggleframe:Hide()
        else
            T.F.toggleframe:Show()
        end
    end

    function BW_ToggleBank(bank) -- Bindings.xml
        local banks = {
            ['bank'] = T.F.mainframe.bankbutton,
            ['reagent'] = T.F.mainframe.reagentbutton,
            ['character'] = T.F.mainframe.characterbutton,
            ['void'] = T.F.mainframe.voidbutton,
            ['guild'] = T.F.mainframe.guildbutton
        }
        T.F.mainframe:Show()
        banks[bank]:Click()
    end

--[[//////////////////////////////////////////////////

    MAIN FUNCTIONS

----\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\]]

    function B.Initialize()

        hooksecurefunc("ContainerFrameItemButton_OnModifiedClick", B.ContainerFrameItemButton_OnModifiedClick)
        hooksecurefunc("ChatFrame_OnHyperlinkShow", B.ChatFrame_OnHyperlinkShow)
        hooksecurefunc("ChatEdit_InsertLink", B.InsertLink)
        hooksecurefunc("SetItemButtonQuality", B.ButtonListener) -- Magic workaround multiple event triggering

        local LibExtraTip = LibStub("LibExtraTip-1")
        LibExtraTip:AddCallback({type = "item", callback = B.TooltipListener})
        LibExtraTip:RegisterTooltip(GameTooltip)

        local player = UnitName('player')
        local server = GetRealmName()
        local faction = UnitFactionGroup('player')

        B.Load(player, server, faction)
    end

    function B.Load(player, server, faction)

        if (player == nil) or (player == UNKNOWNOBJECT) or (player == UNKNOWNBEING) then
            return
        end

        if BWsavedtable and BWsavedtable[server] and BWsavedtable[server][player.."."..string.sub(faction, 1, 1)] then
            B.load(server, player, faction)
        else
            B.print(T.MSGC.EVENT.."New character detected")
            B.template(server, player, faction)
        end

        -- Reference tables for WoW loaded character
        if player == UnitName('player') then
            T['bank'] = B.GetPlayerBank('bank', true)
            T['reagent'] = B.GetPlayerBank('reagent', true)
            T['void'] = B.GetPlayerBank('void', true)
            T['guild'] = B.GetPlayerBank('guild', true)
            T['player'] = B.GetPlayerBank('player', true)
        end
    end

    function B.Main(input)

        B.print(T.MSGC.INFO.."That's a lot of stuff.", 1)
        if input then input = string.lower(input) end

        if input == "toggle" then
            BW_ToggleFrame()

        elseif input == "debug" then
            BWdebug = not BWdebug
            B.print(T.MSGC.INFO.."Debug:|r "..T.MSGC.BOOL..tostring(BWdebug), 1)
        end
    end

    function B.ButtonListener(button, quality, link)
        -- Workaround to using events which may trigger more than once.
        --B.print(button:GetName())
        if B.IsGuildBank(button) then
            B.GuildSlotScan(GetCurrentGuildBankTab(), button:GetID(), button, link)
        
        elseif B.IsPlayerBag(button) then
            B.PlayerBagSlotScan(button:GetParent():GetID(), button:GetID())

        elseif B.IsBankBag(button) then
            B.BankBagSlotScan(button:GetParent():GetID(), button:GetID())
        
        elseif B.IsReagentBank(button) then
            B.ReagentSlotScan(button:GetID())

        elseif B.IsGenericBank(button) then
            B.GenericBankSlotScan(button:GetID())

        elseif B.IsVoidBank(button) then
            B.VoidSlotScan(tonumber(VoidStorageFrame.page), button:GetID(), button)
        end
    end

    function B.TooltipListener(tooltip, itemlink, quantity)
        -- Minimal approach
        local itemid, bank, bag, p, newstring, string, total
        local player = UnitName('player')

        local ignored = {
            [6948] = true, -- Hearthstone
            [110560] = true, -- Garrison Hearthstone
            [140192] = true -- Dalaran Hearthstone
        }

        local order = {"player", "equipped", "bank", "reagent", "void", "guild"}
        local r, g, b = ITEM_QUALITY_COLORS[6].r, ITEM_QUALITY_COLORS[6].g, ITEM_QUALITY_COLORS[6].b --0.74, 0.46, 1.00, nice purple...
        local contrast = ITEM_QUALITY_COLORS[5].hex

        if itemlink then
            itemid = tonumber(B.GetItemIdFromLink(itemlink))
            if ignored[itemid] then return end

            string = B.newtable()
            total = 0

            -- Main player
            for x = 1, #order do
                bank = order[x]
                bag = T.BW.tooltipcache[player][bank]

                if bag then
                    result = bag[itemid]
                    if result then
                        string[#string + 1] = contrast..T.TOOLTIPLABEL[bank].."|r: |cffffffff"..result.."|r"
                        total = total + result
                    end
                end
            end
            tooltip:AddLine(" ")
            if next(string) then tooltip:AddDoubleLine(player..contrast.."|r (|cffffffff "..total.." |r"..contrast.."|r)", "["..table.concat(string, ", ").."]", r, g, b, r, g, b) end

            -- Alts
            table.wipe(string)
            for p, bank in pairs(T.BW.tooltipcache) do
                if p ~= player then
                    for x = 1, #order do
                        bank = order[x]
                        bag = T.BW.tooltipcache[p][bank]

                        if bag then
                            result = bag[itemid]
                            if result then
                                string[#string + 1] = contrast..T.TOOLTIPLABEL[bank].."|r: |cffffffff"..result.."|r"
                            end
                        end
                    end
                    if next(string) then tooltip:AddDoubleLine(p, "["..table.concat(string, ", ").."]", r, g, b, r, g, b) end
                    table.wipe(string)
                end
            end
        end
    end

    function B.ContainerFrameItemButton_OnModifiedClick(self)

        if IsShiftKeyDown() then
            local bag, item = self:GetParent():GetID(), self:GetID()
            local link = GetContainerItemLink(bag, item)

            if T.F.mainframe.searchbar:HasFocus() and T.F.mainframe.searchbar:GetText() ~= text then

                T.F.mainframe.searchbar:SetText(GetItemInfo(link))
            end
        end
    end

    function B.ChatFrame_OnHyperlinkShow(self, link, text, button)

        if IsShiftKeyDown() and link then
            if string.find(link, 'achievement:*') or string.find(link,'spell:*') then
                return false
            end

            if T.F.mainframe.searchbar:HasFocus() and T.F.mainframe.searchbar:GetText() ~= text then
                T.F.mainframe.searchbar:SetText(GetItemInfo(link))
            end
        end
    end

    function B.InsertLink(text)

        if T.F.mainframe.searchbar:HasFocus() and T.F.mainframe.searchbar:GetText() ~= text then
            T.F.mainframe.searchbar:SetText(string.find(text, "^|c%x+|H(.+)|h%[.*%]"))
        end
    end

    function B.dropDownMenuItem_OnClick(self)
        local server, player, faction
        local value = self.value

        if value == "delete" then
            BWsavedtable[T.ACTIVE_TABLE.server][T.ACTIVE_TABLE.player.."."..string.sub(T.ACTIVE_TABLE.faction, 1, 1)] = nil

            local player = UnitName('player')
            local server = GetRealmName()
            local faction = UnitFactionGroup('player')
            B.Load(player, server, faction)

            B.PlayerBagsScan()
            B.PlayerEquipmentScan()
            B.PlayerEquipmentUpdate()

            -- Cache items here
            B.CacheMainItems()
            B.CacheAltItems()
        else
            server, player = value:match"([^.]*).(.*)"
            player, faction = player:match"([^.]*).(.*)"
            faction = faction == "H" and "Horde" or faction == "A" and "Alliance" or faction == "N" and "Neutral"
            B.load(server, player, faction)
        end

        CloseDropDownMenus()
        --BWdropDownMenuButton:Click()
    end

--[[//////////////////////////////////////////////////

    SCAN FUNCTIONS (OLDER)

----\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\]]
    
    function B.BankScan()

        if not T.ISBANKOPEN then return end -- Circumvent early BAG_UPDATE
        local x, counter, count, link, bag, slot, slots
        local bank = T['bank']

        if not bank[1] then bank[1] = B.newtable() end
        for x = 1, NUM_BANKGENERIC_SLOTS do
            _, count, _, _, _, _, link = GetContainerItemInfo(BANK_CONTAINER, x)

            if link then
                bank[1][x] = bank[1][x] or B.newtable()
                bank[1][x]['count'] = count
                bank[1][x]['link'] = link
            else
                bank[1][x] = B.deltable(bank[1][x])
            end
        end

        for bag = ITEM_INVENTORY_BANK_BAG_OFFSET + 1, ITEM_INVENTORY_BANK_BAG_OFFSET + NUM_BANKBAGSLOTS do
            if not bank[bag] then bank[bag] = B.newtable() end
            
            local inventoryID = BankButtonIDToInvSlotID(bag, 1) - 4
            local baglink = GetInventoryItemLink("player", inventoryID)

            slots = GetContainerNumSlots(bag)
            bank[bag]['slots'] = slots
            bank[bag]['link'] = baglink

            for x = 1, slots do
                _, count, _, _, _, _, link = GetContainerItemInfo(bag, x)

                if link then
                    bank[bag][x] = bank[bag][x] or B.newtable()
                    bank[bag][x]['link'] = link
                    bank[bag][x]['count'] = count > 1 and count or nil
                else
                    bank[bag][x] = B.deltable(bank[bag][x])
                end
            end
        end
    end

    function B.ReagentScan()

        local x, page, itemID
        local reagent = T['reagent']

        for page = 1, T.REAGENT_PAGES do
            if not reagent[page] then reagent[page] = B.newtable() end

            for x = 1, T.NUM_REAGENT_SLOTS do

                local _, count, _, _, _, _, link = GetContainerItemInfo(REAGENTBANK_CONTAINER, slot)

                if link then
                    reagent[page][x] = reagent[page][x] or B.newtable()
                    reagent[page][x]['link'] = link
                    reagent[page][x]['count'] = count > 1 and count or nil
                else
                    reagent[page][x] = B.deltable(reagent[page][x])
                end
            end
        end
    end

    function B.VoidScan()

        if not IsVoidStorageReady() then return end
        local x, page, itemID
        local void = T['void']

        for page = 1, T.VOID_PAGES do
            if not void[page] then void[page] = B.newtable() end

            for x = 1, T.NUM_VOID_SLOTS do
                itemID = GetVoidItemInfo(page, x)

                if itemID then
                    void[page][x] = void[page][x] or B.newtable()
                    void[page][x]['link'] = page == 1 and GetVoidItemHyperlinkString(x) or GetVoidItemHyperlinkString(T.NUM_VOID_SLOTS + x)
                else
                    void[page][x] = B.deltable(void[page][x])
                end
            end
        end
    end

    function B.GuildScan()

        local x, i, id, count, item, button
        local guild = T['guild']
        local page = GetCurrentGuildBankTab()

        if guild[page].access then
            for x = 1, T.MAX_GUILDBANK_SLOTS_PER_TAB do

                _, count = GetGuildBankItemInfo(page, x)
                if count > 0 then
                    guild[page][x] = guild[page][x] or B.newtable()
                    guild[page][x]['link'] = GetGuildBankItemLink(page, x)
                    guild[page][x]['count'] = count > 1 and count or nil
                else
                    guild[page][x] = B.deltable(guild[page][x])
                end
            end
        end
    end

--[[//////////////////////////////////////////////////

    SCAN FUNCTIONS

----\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\]]
    
    function B.Scan(slot, link, count)

        if link then
            slot = slot or B.newtable()
            slot['link'] = link
            slot['count'] = count and count > 1 and count or nil
        else
            slot = B.deltable(slot)
        end

        return slot
    end

    function B.BankBagScan(bag)

        local x, slots, count, link
        local bank = T['bank']

        B.RefreshBankBag(bank, bag)

        for x = 1, bank[bag]['slots'] do
            _, count, _, _, _, _, link = GetContainerItemInfo(bag, x)
            bank[bag][x] = B.Scan(bank[bag][x], link, count)
        end

        if B.IsCurrentPlayer() then B.BankUpdate() end
    end

    function B.GenericBankSlotScan(slot)

        local bank = T['bank']
        local _, count, _, _, _, _, link = GetContainerItemInfo(BANK_CONTAINER, slot)

        bank[1][slot] = B.Scan(bank[1][slot], link, count)
        if B.IsCurrentPlayer() then B.GenericBankSlotUpdate(slot) end
    end

    function B.BankBagSlotScan(bag, slot)

        local bank = T['bank']
        local _, count, _, _, _, _, link = GetContainerItemInfo(bag, slot)

        B.RefreshBankBag(bank, bag)

        bank[bag][slot] = B.Scan(bank[bag][slot], link, count)
        if B.IsCurrentPlayer() then B.BankSlotUpdate(B.GetSlotID(bag, slot), bag, slot) end
    end

    function B.ReagentSlotScan(slot)

        local reagent = T['reagent']
        local _, count, _, _, _, _, link = GetContainerItemInfo(REAGENTBANK_CONTAINER, slot)

        if not reagent[1] then reagent[1] = B.newtable() end

        reagent[1][slot] = B.Scan(reagent[1][slot], link, count)
        if B.IsCurrentPlayer() then B.ReagentSlotUpdate(slot) end
    end

    function B.VoidSlotScan(page, slot, button)

        local void = T['void']

        if not void[page] then void[page] = B.newtable() end

        if button.hasItem then
            void[page][slot] = void[page][slot] or B.newtable()
            void[page][slot]['link'] = page == 1 and GetVoidItemHyperlinkString(slot) or GetVoidItemHyperlinkString(T.NUM_VOID_SLOTS + slot)
        else
            void[page][slot] = B.deltable(void[page][slot])
        end

        if B.IsCurrentPlayer() then B.VoidSlotUpdate(page, slot) end
    end

    function B.GuildSlotScan(page, slot, button, link)

        local guild = T['guild']

        if button.count > 0 then
            if strfind(link, "item:82800:") then -- Caged pet
                link = B.GetGuildPetLink(page, slot)
            end
        end

        guild[page][slot] = B.Scan(guild[page][slot], link, button.count)
        if B.IsCurrentPlayer() then B.GuildSlotUpdate(page, slot) end
    end

    function B.PlayerBagsScan()

        local x, counter, count, link, bag, slot, slots
        local player = T['player']

        for bag = BACKPACK_CONTAINER, NUM_BAG_SLOTS do
            if not player[bag] then player[bag] = B.newtable() end

            slots = GetContainerNumSlots(bag)
            player[bag]['slots'] = slots

            if bag > 0 then
                local inventoryID = ContainerIDToInventoryID(bag)
                local baglink = GetInventoryItemLink("player", inventoryID)
                player[bag]['link'] = baglink
            end

            for x = 1, slots do
                _, count, _, _, _, _, link = GetContainerItemInfo(bag, x)

                player[bag][x] = B.Scan(player[bag][x], link, count)
            end
        end   
    end

    function B.PlayerBagScan(bag)

        local x, slots, count, link
        local player = T['player']

        if not player[bag] then player[bag] = B.newtable() end
            
        slots = GetContainerNumSlots(bag)
        player[bag]['slots'] = slots

        if bag > 0 then
            local inventoryID = ContainerIDToInventoryID(bag)
            local baglink = GetInventoryItemLink("player", inventoryID)
            player[bag]['link'] = baglink
        end

        for x = 1, slots do
            _, count, _, _, _, _, link = GetContainerItemInfo(bag, x)
            player[bag][x] = B.Scan(player[bag][x], link, count)
        end

        if B.IsCurrentPlayer() then B.PlayerUpdate() end
    end

    function B.PlayerBagSlotScan(bag, slot)

        local player = T['player']
        local _, count, _, _, _, _, link = GetContainerItemInfo(bag, slot)

        if not player[bag] then
            player[bag] = B.newtable()
            slots = GetContainerNumSlots(bag)
            player[bag]['slots'] = slots

            if bag > 0 then
                local inventoryID = ContainerIDToInventoryID(bag)
                local baglink = GetInventoryItemLink("player", inventoryID)
                player[bag]['link'] = baglink
            end
        end

        player[bag][slot] = B.Scan(player[bag][slot], link, count)
        if B.IsCurrentPlayer() then B.PlayerSlotUpdate(bag, slot) end
    end

    function B.PlayerEquipmentScan()

        local inventory, slotId, link
        local player = T['player']

        if not player[-1] then player[-1] = B.newtable() end

        -- Tabard slot is slotId 19?
        for x = 1, #T.INVSLOTS do

            slotId = GetInventorySlotInfo(T.INVSLOTS[x])
            link = GetInventoryItemLink("player", slotId)

            if slotId == 19 then slotId = 18 end

            player[-1][slotId] = B.Scan(player[-1][slotId], link)
        end
    end

    function B.PlayerEquipmentSlotScan(slot)

        local player = T['player']
        local link = GetInventoryItemLink("player", slot)

        if slot == 19 then slot = 18 end

        player[-1][slot] = B.Scan(player[-1][slot], link)
        if B.IsCurrentPlayer() then B.PlayerEquipmentSlotUpdate(slot) end
    end

    function B.BankOfflineScan(bag)
        -- Item count is always updated. Check the difference for the bag.
        local slot, count, total
        local bank = T['bank']

        total = 0
        for x = 1, bank[bag]['slots'] do
            slot = bank[bag][x]
            if slot and next(slot) then
                count = (GetItemCount(slot['link'], true) or 0) - (GetItemInfo(slot['link']) or 0)
                B.print(slot['link'])
                B.print(count)
                --[[if count then -- item just changed
                    B.print(slot['link'])
                    B.print(count)
                    B.print(slot['count'])
                end]]
            end
        end
    end

--[[//////////////////////////////////////////////////

    UPDATE FUNCTIONS

----\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\]]
    
    function B.ScrollbarUpdate()
        local r, g, b, idx, link, hexcolor, itemRarity
        local total_items = #T.BW.items
        local currentscrollbarvalue = FauxScrollFrame_GetOffset(T.F.mainframe.search.scroll)

        FauxScrollFrame_Update(T.F.mainframe.search.scroll, total_items, B.RowsToFit(), T.BUTTON_HEIGHT)
        
        for x = 1, B.RowsToFit() do
            -- Clear buttons
            B.ResetButton(T.F.mainframe.search.list.buttons[x])
            T.F.mainframe.search.list.buttons[x]:SetBackdropColor(0, 0, 0, 0.7)
            T.F.mainframe.search.list.buttons[x].leftstring:SetText("")
            T.F.mainframe.search.list.buttons[x]:Hide()
            T.F.mainframe.search.enddot:Hide()
        end

        for x = 1, B.RowsToFit() do
            -- Get the appropriate item, which will be x + value
            if (T.F.mainframe.layout == 2 or T.F.mainframe.layout == 4) and total_items > 0 and total_items < B.RowsToFit() then
                local difference = B.RowsToFit() - total_items
                buttonnum = x + difference
            else
                buttonnum = x
            end

            idx = x + currentscrollbarvalue

            if T.BW.items[idx] then
                r, g, b, hexcolor = GetItemQualityColor(T.BW.items[idx].quality)

                T.F.mainframe.search.list.buttons[buttonnum].quest = T.BW.items[idx].quest
                T.F.mainframe.search.list.buttons[buttonnum].quality = T.BW.items[idx].quality
                T.F.mainframe.search.list.buttons[buttonnum].icon:SetTexture(T.BW.items[idx].icon)
                T.F.mainframe.search.list.buttons[buttonnum].leftstring:SetText("|c"..hexcolor..T.BW.items[idx].name)

                if T.BW.items[idx].count > 1 then
                    T.F.mainframe.search.list.buttons[buttonnum].text:SetText(T.BW.items[idx].count)
                end

                if T.BW.items[idx].ilvl and T.BW.items[idx].ilvl ~= "" then
                    T.F.mainframe.search.list.buttons[buttonnum].ilvl:SetText(T.BW.items[idx].ilvl)
                    T.F.mainframe.search.list.buttons[buttonnum].ilvl:SetTextColor(r, g, b, 1)
                end

                --loc, num = T.BW.items[idx].loc:match"([^.]*).(.*)"
                if T.BW.items[idx].loc == "alt" then
                    location = T.BW.items[idx].num
                else
                    location = T.BW.items[idx].num == 1 and T.LABEL[T.BW.items[idx].loc] or T.LABEL[T.BW.items[idx].loc].." "..T.BW.items[idx].num
                end
                T.F.mainframe.search.list.buttons[buttonnum].location = {['loc'] = T.BW.items[idx].loc, ['num'] = T.BW.items[idx].num}
                T.F.mainframe.search.list.buttons[buttonnum].rightstring:SetText(location)
                T.F.mainframe.search.list.buttons[buttonnum].link = T.BW.items[idx].link
                T.F.mainframe.search.list.buttons[buttonnum]:Show()
                T.F.mainframe.search.enddot:Show()
                if idx == total_items then
                    T.F.mainframe.search.enddot:Hide()
                end
            end
        end
    end

    function B.BankUpdate()

        local page, total, counter, quality, r, g, b
        local bank = B.GetPlayerBank('bank')

        B.ValidateBankBags(bank)

        counter = 0
        for page = 1, NUM_BAG_SLOTS + NUM_BANKBAGSLOTS do
            if bank[page] then

                if page > 1 then -- Ignore generic bank
                    B.UpdateBagTab(T.F.mainframe.bank.bags[page - 4].tab, bank[page])
                end
            
                B.UpdatePage(T.F.mainframe.bank, bank[page]['slots'], bank[page], counter)
                counter = counter + bank[page]['slots']
            
            elseif page >= NUM_BAG_SLOTS + 1 then
                B.UpdateBagTab(T.F.mainframe.bank.bags[page - 4].tab, {})
            end
        end
    end

    function B.PlayerUpdate()

        local page, total, counter, quality, r, g, b
        local player = B.GetPlayerBank('player')

        B.ValidatePlayerBags(player)

        counter = 0
        for page = 0, #player do

            if page > 0 then -- Ignore generic bag
                B.UpdateBagTab(T.F.mainframe.character.bags[page].tab, player[page])
            end

            if player[page] then

                B.UpdatePage(T.F.mainframe.character, player[page]['slots'], player[page], counter)
                counter = counter + player[page]['slots']
            end
        end
    end

    function B.PlayerSlotUpdate(bag, slot)

        local page, counter, item
        local player = B.GetPlayerBank('player')

        B.ValidatePlayerBags(player)

        counter = 0
        for page = 0, #player do

            if page > 0 then -- Ignore generic bag
                B.UpdateBagTab(T.F.mainframe.character.bags[page].tab, player[page])
            end

            if player[page] then
                if page == bag then
                    
                    B.ResetButton(T.F.mainframe.character.buttons[counter + slot])
                    if player[bag][slot] then

                        item = B.GetItem(player[bag][slot].link)
                        if next(item) then B.UpdateButton(T.F.mainframe.character.buttons[counter + slot], item, player[bag][slot]) end
                    end
                end
                counter = counter + player[page]['slots']
            end
        end
    end

    function B.GenericBankSlotUpdate(slot)
        local bank = B.GetPlayerBank('bank')
        B.UpdateSlot(T.F.mainframe.bank.buttons[slot], bank[1][slot])
    end

    function B.BankSlotUpdate(button, page, slot)
        local bank = B.GetPlayerBank('bank')
        B.UpdateSlot(T.F.mainframe.bank.buttons[button], bank[page][slot])
    end

    function B.PlayerEquipmentUpdate()
        local player = B.GetPlayerBank('player')
        B.UpdatePage(T.F.mainframe.character.equipment, 18, player[-1])
    end

    function B.PlayerEquipmentSlotUpdate(slot)
        local player = B.GetPlayerBank('player')
        B.UpdateSlot(T.F.mainframe.character.equipment.buttons[slot], player[-1][slot])
    end

    function B.ReagentSlotUpdate(slot)
        local reagent = B.GetPlayerBank('reagent')
        B.UpdateSlot(T.F.mainframe.reagent[1].buttons[slot], reagent[1][slot])
    end

    function B.VoidSlotUpdate(page, slot)
        local void = B.GetPlayerBank('void')
        B.UpdateSlot(T.F.mainframe.void[page].buttons[slot], void[page][slot])
    end

    function B.GuildSlotUpdate(page, slot)
        local guild = B.GetPlayerBank('guild')
        B.UpdateSlot(T.F.mainframe.guild[page].buttons[slot], guild[page][slot])       
    end

    function B.ReagentUpdate()
        local page, reagent = _, B.GetPlayerBank('reagent')

        for page = 1, T.REAGENT_PAGES do
            B.UpdatePage(T.F.mainframe.reagent[page], T.NUM_REAGENT_SLOTS, reagent[page])
        end
    end

    function B.VoidUpdate()
        local page, void = _, B.GetPlayerBank('void')

        for page = 1, T.VOID_PAGES do
            B.UpdatePage(T.F.mainframe.void[page], T.NUM_VOID_SLOTS, void[page])
        end
    end

    function B.GuildUpdate()

        local page, guild = _, B.GetPlayerBank('guild')

        for page = 1, MAX_GUILDBANK_TABS do
            if guild[page] then

                if guild[page].icon then
                    T.F.mainframe.guild[page].tab:SetNormalTexture(guild[page].icon)
                    T.F.mainframe.guild[page].tab.name = guild[page].name
                    T.F.mainframe.guild[page].tab:Show()
                else
                    T.F.mainframe.guild[page].tab:Hide()
                end

                if guild[page].access then -- Validate user has access to view tab
                    T.F.mainframe.guild[page].tab:Enable()
                    T.F.mainframe.guild[page].tab:GetNormalTexture():SetDesaturated(false)
                    B.UpdatePage(T.F.mainframe.guild[page], T.MAX_GUILDBANK_SLOTS_PER_TAB, guild[page])
                else
                    T.F.mainframe.guild[page].tab:Disable()
                    T.F.mainframe.guild[page].tab:GetNormalTexture():SetDesaturated(true)
                end
            else
                T.F.mainframe.guild[page].tab:Hide()
                B.UpdatePage(T.F.mainframe.guild[page], T.MAX_GUILDBANK_SLOTS_PER_TAB, nil)
            end
         end
    end

    function B.RefreshGuildTab()

        local x, name, icon, isViewable
        local guild = T['guild']
        local tabs = GetNumGuildBankTabs()

        if IsInGuild() then guild['name'] = GetGuildInfo('player') end

        for x = 1, MAX_GUILDBANK_TABS do
            if x <= tabs then
                if not guild[x] then guild[x] = B.newtable() end
                name, icon, isViewable = GetGuildBankTabInfo(x)
                guild[x]['icon'] = icon
                guild[x]['name'] = name
                guild[x]['access'] = isViewable

                T.F.mainframe.guild[x].tab:SetNormalTexture(icon)
                T.F.mainframe.guild[x].tab.name = name
                T.F.mainframe.guild[x].tab:Enable()
                T.F.mainframe.guild[x].tab:GetNormalTexture():SetDesaturated(false)
                T.F.mainframe.guild[x].tab:Show()
            else
                guild[x] = B.deltable(guild[x])
                T.F.mainframe.guild[x].tab:Disable()
                T.F.mainframe.guild[x].tab:GetNormalTexture():SetDesaturated(true)
                T.F.mainframe.guild[x].tab:Hide()
            end
        end
    end

    function B.RefreshBankBags()
        local total, page
        local bank = T['bank']

        for bag = ITEM_INVENTORY_BANK_BAG_OFFSET + 1, ITEM_INVENTORY_BANK_BAG_OFFSET + NUM_BANKBAGSLOTS do
            B.RefreshBankBag(bank, bag)

            if bag > 1 then -- Ignore generic bank
                B.UpdateBagTab(T.F.mainframe.bank.bags[bag - 4].tab, bank[bag])
            end
        end

        B.ValidateBankBags(bank)
    end

    function B.RefreshBankBag(bank, bag)

        if not bank[bag] then bank[bag] = B.newtable() end
            
        local inventoryID = BankButtonIDToInvSlotID(bag, 1) - 4
        local baglink = GetInventoryItemLink("player", inventoryID)
        local slots = GetContainerNumSlots(bag)

        bank[bag]['slots'] = slots
        bank[bag]['link'] = baglink

        B.ValidateBankBags(bank)
    end

    function B.RefreshPlayerBag(player, bag)
        local slots

        if not player[bag] then
            player[bag] = B.newtable()
            slots = GetContainerNumSlots(bag)
            player[bag]['slots'] = slots

            if bag > 0 then
                local inventoryID = ContainerIDToInventoryID(bag)
                local baglink = GetInventoryItemLink("player", inventoryID)
                player[bag]['link'] = baglink
            end
        end
    end

    function B.UpdateBagTab(tab, data)
        local quality, r, g, b

        if data.link then
            quality = B.GetQualityFromLink(data.link)
            r, g, b = GetItemQualityColor(quality)

            tab:SetBackdropBorderColor(r, g, b, 1)
            tab.link = data.link
            tab.quality = quality
            tab:SetNormalTexture(GetItemIcon(data.link))
            tab:Show()
        else
            tab:Hide()
        end
    end

    function B.UpdatePage(page, maxslots, data, optstart)
        local x, item
        optstart = optstart or 0

        for x = 1, maxslots do
            B.UpdateSlot(page.buttons[x + optstart], data and data[x] or nil)
        end
    end

    function B.UpdateSlot(button, data)

        if button then
            B.ResetButton(button)

            if data then
                item = B.GetItem(data.link)
                if next(item) then B.UpdateButton(button, item, data) end
            end
        end
    end

    local item = {}
    function B.GetItem(itemlink)

        local name, link, quality, iLevel, reqLevel, class, subclass, maxStack, equipSlot, texture, vendorPrice

        if strmatch(itemlink, "|Hbattlepet:") then
            name, link, quality, iLevel, reqLevel, class, subclass, maxStack, equipSlot, texture, vendorPrice = B.GetPetInfo(itemlink)
        else
            name, link, quality, iLevel, reqLevel, class, subclass, maxStack, equipSlot, texture, vendorPrice = GetItemInfo(itemlink)
        end

        -- Reuse the same table to avoid generating garbage
        item[1] = name
        item[2] = link
        item[3] = quality
        item[4] = iLevel
        item[5] = reqLevel
        item[6] = class
        item[7] = subclass
        item[8] = maxStack
        item[9] = equipSlot
        item[10] = texture
        item[11] = vendorPrice

        if not next(item) then -- Item is not cached yet
            T.BW.lookuprequest = true
            T.BW.lookup[tonumber(B.GetItemIdFromLink(itemlink))] = true
        end

        return item
    end

    function B.UpdateButton(button, item, data)
        local r, g, b, ilvl

        r, g, b = GetItemQualityColor(item[3])

        button.name = item[1]
        button.count = data.count or 1
        button.quality = item[3]
        button.link = item[2]
        button.class = item[6]
        button.subclass = item[7]
        button.equipslot = item[9]

        button.icon:SetTexture(item[10])
        button.icon:SetTexCoord(0.1, 0.9, 0.1, 0.9)
        
        if (data.count or 1) > 1 then button.text:SetText(data.count) end
        if item[3] > 1 then button:SetBackdropBorderColor(r, g, b, 1) end

        if item[6] == BATTLE_PET_SOURCE_2 then -- QUEST
            button:SetBackdropBorderColor(1, 1, 0, 1)
            button.quest = true

        elseif item[6] == ARMOR or item[6] == ENCHSLOT_WEAPON then -- ILVL
            ilvl = B.DetectILVL(item[2])
            if ilvl then
                button.ilvl:SetText(ilvl)
                button.ilvl:SetTextColor(r, g, b, 1)
            else
                button.ilvl:SetText(item[4])
                button.ilvl:SetTextColor(r, g, b, 1)
            end
        end
    end

    function B.ResetButton(button)

        button.icon:SetTexture("")
        button:SetBackdropBorderColor(0, 0, 0, 1)
        button.link = nil
        button.quality = nil
        button.quest = nil
        button.name = nil
        button.count = nil
        button.text:SetText("")
        button.ilvl:SetText("")
    end

--[[//////////////////////////////////////////////////

    SEARCH FUNCTIONS

----\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\]]

    function B.Search(text)

        if text == "" or text == SEARCH then text = false end

        T.F.mainframe.search.scroll.ScrollBar:SetValue(0)

        B.deltable(T.BW.items)
        T.BW.items = B.newsearchtable()

        B.BankSearch(text)
        B.ReagentSearch(text)
        B.VoidSearch(text)
        B.GuildSearch(text)
        B.PlayerSearch(text)
        B.PlayerEquipmentSearch(text)
        B.SearchAlts(text)

        B.ScrollbarUpdate()

        if text and T.F.mainframe.countbutton.count > 0 then
            T.F.mainframe.countbutton.labelcount:SetText("|cffffffff"..T.F.mainframe.countbutton.count.."|r")
            T.F.mainframe.countbutton:Show()
            if not T.F.mainframe.search:IsVisible() and not T.F.mainframe.lastbank then
                T.F.mainframe.countbutton:Click()
            end
        else
            T.F.mainframe.countbutton:Hide()
        end
    end

    function B.BankSearch(text)

        local x, page, counter, button, buttons, name
        local bank = B.GetPlayerBank('bank')

        local counter = 1
        for page = 1, #bank do
            if bank[page] then

                if page > 1 then -- ignore generic bank
                    local tab = page - 4

                    if text then T.F.mainframe.bank.bags[tab].tab.search:Show()
                    else T.F.mainframe.bank.bags[tab].tab.search:Hide() end
                end

                buttons = T.F.mainframe.bank.buttons

                for x = 1, bank[page]['slots'] do
                    button = buttons[counter]

                    if button.name and text then
                        if B.find(text, button) then
                            button.loc = "bank."..(page == 1 and page or page - 4)
                            B.SearchSlot(button)
                            if page > 1 then local tab = page - 4; T.F.mainframe.bank.bags[tab].tab.search:Hide() end
                        else
                            button.search:Show()
                        end
                    else
                        button.search:Hide()
                    end
                    counter = counter + 1
                end
            end
        end
    end

    function B.PlayerSearch(text)

        local x, page, counter, button, buttons, name
        local player = B.GetPlayerBank('player')

        local counter = 1
        for page = 0, #player do
            if player[page] then

                if page > 0 then -- ignore generic bag
                    local tab = page

                    if text then T.F.mainframe.character.bags[tab].tab.search:Show()
                    else T.F.mainframe.character.bags[tab].tab.search:Hide() end
                end

                buttons = T.F.mainframe.character.buttons
                for x = 1, player[page]['slots'] do
                    button = buttons[counter]

                    if button.name and text then
                        if B.find(text, button) then
                            button.loc = "player."..(page == 0 and 1 or page + 1)
                            B.SearchSlot(button)
                            if page > 0 then local tab = page; T.F.mainframe.character.bags[tab].tab.search:Hide() end
                        else
                            button.search:Show()
                        end
                    else
                        button.search:Hide()
                    end
                    counter = counter + 1
                end
            end
        end
    end

    function B.PlayerArtifactSearch()

        local x, page, counter, button, buttons, name
        local player = T['player']

        local counter = 1
        for page = 0, #player do
            if player[page] then

                if page > 0 then -- ignore generic bag
                    local tab = page

                    if text then T.F.mainframe.character.bags[tab].tab.search:Show()
                    else T.F.mainframe.character.bags[tab].tab.search:Hide() end
                end

                buttons = T.F.mainframe.character.buttons
                for x = 1, player[page]['slots'] do
                    button = buttons[counter]

                    if button.name and text then
                        if B.find(strlower(ARTIFACT_POWER), button) then
                            if not T.F.mainframe.character.artifact.total then T.F.mainframe.character.artifact.total = B.DetectAP(B.GetBag(button)) end
                            if page > 0 then local tab = page; T.F.mainframe.character.bags[tab].tab.search:Hide() end
                        else
                            button.search:Show()
                        end
                    else
                        button.search:Hide()
                    end
                    counter = counter + 1
                end
            end
        end
    end

    function B.PlayerEquipmentSearch(text)

        local x, button, buttons, name
        local player = B.GetPlayerBank('player')

        if not player[-1] then return end

        buttons = T.F.mainframe.character.equipment.buttons
        for x = 1, #player[-1] do
            button = buttons[x]

            if button.name and text then
                if B.find(text, button) then
                    button.loc = "equip.1"
                    B.SearchSlot(button)
                else
                    button.search:Show()
                end
            else
                button.search:Hide()
            end

        end
    end

    function B.BankBagSearch(bag)

        local bank = B.GetPlayerBank('bank')
        local x, page, counter, button, buttons

        local counter = 1
        for page = 1, #bank do
            if bank[page] then

                if not bag or page == bag then
                    for x = 1, bank[page]['slots'] do
                        T.F.mainframe.bank.buttons[counter].search:Hide()
                        counter = counter + 1
                    end
                else
                    for x = 1, bank[page]['slots'] do
                        T.F.mainframe.bank.buttons[counter].search:Show()
                        counter = counter + 1
                    end
                end
            end
        end
    end

    function B.PlayerBagSearch(bag)

        local player = B.GetPlayerBank('player')
        local x, page, counter, button, buttons

        local counter = 1
        for page = 0, #player do
            if player[page] then

                if not bag or page == bag then
                    for x = 1, player[page]['slots'] do
                        T.F.mainframe.character.buttons[counter].search:Hide()
                        counter = counter + 1
                    end
                else
                    for x = 1, player[page]['slots'] do
                        T.F.mainframe.character.buttons[counter].search:Show()
                        counter = counter + 1
                    end
                end
            end
        end
    end

    function B.ReagentSearch(text)

        local x, page, button, buttons
        local reagent = B.GetPlayerBank('reagent')

        for page = 1, T.REAGENT_PAGES do
            buttons = T.F.mainframe.reagent[page].buttons
            for x = 1, T.NUM_REAGENT_SLOTS do
                button = buttons[x]

                if button.name and text then
                    if B.find(text, button) then
                        button.loc = "reagent."..page
                        B.SearchSlot(button)
                    else
                        button.search:Show()
                    end
                else
                    button.search:Hide()
                end
            end
        end
    end

    function B.VoidSearch(text)

        local x, page, button, buttons
        local void = B.GetPlayerBank('void')

        for page = 1, T.VOID_PAGES do

            if text then T.F.mainframe.void[page].tab.search:Show() 
            else T.F.mainframe.void[page].tab.search:Hide() end

            buttons = T.F.mainframe.void[page].buttons

            for x = 1, T.NUM_VOID_SLOTS do
                button = buttons[x]

                if button.name and text then
                    if B.find(text, button) then
                        button.loc = "void."..page
                        B.SearchSlot(button)
                        T.F.mainframe.void[page].tab.search:Hide()
                    else
                        button.search:Show()
                    end
                else
                    button.search:Hide()
                end
            end
        end
    end

    function B.GuildSearch(text)

        local x, page, button, buttons
        local guild = B.GetPlayerBank('guild')

        for page = 1, MAX_GUILDBANK_TABS do

            if text then T.F.mainframe.guild[page].tab.search:Show()
            else T.F.mainframe.guild[page].tab.search:Hide() end

            if guild[page] and guild[page].access then

                buttons = T.F.mainframe.guild[page].buttons

                for x = 1, T.MAX_GUILDBANK_SLOTS_PER_TAB do
                    button = buttons[x]

                    if button.name and text then
                        if B.find(text, button) then
                            button.loc = "guild."..page
                            B.SearchSlot(button)
                            T.F.mainframe.guild[page].tab.search:Hide()
                        else
                            button.search:Show()
                        end
                    else
                        button.search:Hide()
                    end
                end
            end
        end
    end

    function B.SearchAlts(text)
        local p, b, player, faction, item, link, info, temp, bags, bank
        local player, faction, loc, num, name, link, item
        -- Only same server and faction as selected player
        if not text then return end

        for p, b in pairs(BWsavedtable[T.ACTIVE_TABLE.server]) do
            player, faction = p:match"([^.]*).(.*)"

            if faction == string.sub(T.ACTIVE_TABLE.faction, 1, 1) and player ~= T.ACTIVE_TABLE.player then
                for _, bank in pairs(b) do-- banks, bags
                    for _, bags in pairs(bank) do -- bag index, bags
                        if type(bags) == "table" then
                            for _, item in pairs(bags) do
                                if type(item) == "table" then

                                    link = item['link']
                                    info = strmatch(link, "|Hbattlepet:") and {B.GetPetInfo(link)} or {GetItemInfo(link)}

                                    temp = B.newtable()
                                    temp.name = info[1]
                                    temp.quality = info[3]
                                    temp.link = link
                                    temp.class = info[6]
                                    temp.subclass = info[7]
                                    temp.equipslot = info[9]

                                    if B.find(text, temp) then
                                        B.SearchAltSlot(player, info, item['count'] or 1)
                                    end

                                    B.deltable(temp)
                                end
                            end
                        end
                    end
                end
            end
        end
    end

    function B.SearchSlot(button)

        button.search:Hide()

        loc, num = button.loc:match"([^.]*).(.*)"
        num = tonumber(num)

        result = T.BW.items(loc, num, button.name) -- use metatable __call to group items
        if not next(result) then -- New entry
            result['loc'] = loc
            result['num'] = num
            result['name'] = button.name
            result['link'] = button.link
            result['quest'] = button.quest
            result['quality'] = button.quality
            result['ilvl'] = button.ilvl:GetText()
            result['icon'] = button.icon:GetTexture()
            result['count'] = 0
        end

        result['count'] = result['count'] + button.count
        T.F.mainframe.countbutton.count = T.F.mainframe.countbutton.count + button.count
    end

    function B.SearchAltSlot(alt, item, count)

        result = T.BW.items("alt", alt, item[1]) -- use metatable __call to group items
        if not next(result) then -- New entry
            result['loc'] = "alt"
            result['num'] = alt
            result['name'] = item[1]
            result['link'] = item[2]
            result['quality'] = item[3]
            result['quest'] = item[6] == BATTLE_PET_SOURCE_2
            result['ilvl'] = item[6] == ARMOR or item[6] == ENCHSLOT_WEAPON and item[4] or nil
            result['icon'] = item[10]
            result['count'] = 0
        end

        result['count'] = result['count'] + count
    end

--[[//////////////////////////////////////////////////

    TOOLTIP CACHE FUNCTIONS

----\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\]]

    function B.CacheItems(cache, itemid, count)
        if not itemid then return end -- Failsafe (don't know how it's possible thought)

        itemid = tonumber(itemid)

        if not cache[itemid] then
            cache[itemid] = 0
        end

        cache[itemid] = cache[itemid] + count
    end

    function B._CacheItems(player, location, bank)
        local bags, item, b

        T.BW.tooltipcache[player][location] = B.deltable(T.BW.tooltipcache[player][location])
        T.BW.tooltipcache[player][location] = B.newtable()

        if location == "equipped" then
            for _, item in pairs(bank) do
                if type(item) == "table" then

                    B.CacheItems(T.BW.tooltipcache[player][location], B.GetItemIdFromLink(item['link']), item['count'] or 1)
                end
            end
        else
            for b, bags in pairs(bank) do
                if (location == "player" and b > -1) or location ~= "player" then
                    if type(bags) == "table" then
                        for _, item in pairs(bags) do
                            if type(item) == "table" then

                                B.CacheItems(T.BW.tooltipcache[player][location], B.GetItemIdFromLink(item['link']), item['count'] or 1)
                            end
                        end
                    end
                end
            end
        end
    end

    function B.CacheMainItems(bank)
        -- Cache everything from active player into a workable table by itemid - count only.
        local banks, selected
        local player = UnitName('player')

        T.BW.tooltipcache[player] =  T.BW.tooltipcache[player] or B.newtable()

        if not bank then -- Cache everything
            local banks = {"bank", "reagent", "void", "guild", "player", "equipped"}
            for bank = 1, #banks do
                selected = banks[bank] == "equipped" and T['player'][-1] or T[banks[bank]]
                B._CacheItems(player, banks[bank], selected)
            end
        else
            selected = bank == "equipped" and T['player'][-1] or T[bank]
            B._CacheItems(player, bank, selected)
        end
    end

    function B.CacheAltItems() -- Run only once
        local player, faction, p, b, banks, bank, selected

        for p, b in pairs(BWsavedtable[GetRealmName()]) do
            player, faction = p:match"([^.]*).(.*)"

            if faction == string.sub(UnitFactionGroup('player'), 1, 1) and player ~= UnitName('player') then
                T.BW.tooltipcache[player] = B.newtable()

                banks = {"bank", "reagent", "void", "guild", "player", "equipped"}
                for bank = 1, #banks do

                    selected = banks[bank] == "equipped" and b['player'][-1] or b[banks[bank]]
                    B._CacheItems(player, banks[bank], selected)
                end
            end
        end
    end
